"""Unit tests for Postgres backend."""

import time

import pytest
import psycopg2
from sifts.core import SearchEnginePostgreSQL
from psycopg2 import OperationalError

TEST_DB_DSN = "postgresql://testuser:testpass@localhost:5432/testdb"


@pytest.fixture(scope="session")
def docker_compose_file(pytestconfig):
    return "docker-compose.yml"


@pytest.fixture(scope="session")
def postgres_service(docker_services):
    """Wait for the PostgreSQL service to be up and running."""

    def is_postgres_healthy():
        try:
            conn = psycopg2.connect(
                dbname="testdb",
                user="testuser",
                password="testpass",
                host="localhost",
                port=5432,
            )
            conn.close()
            return True
        except OperationalError:
            return False

    timeout = 30
    pause = 0.5
    start_time = time.time()

    while time.time() - start_time < timeout:
        if is_postgres_healthy():
            return
        time.sleep(pause)

    raise TimeoutError("PostgreSQL service did not become healthy in time")


@pytest.fixture
def search_engine():
    engine = SearchEnginePostgreSQL(dsn=TEST_DB_DSN)
    yield engine
    with engine.conn() as conn:
        conn.execute("TRUNCATE TABLE documents RESTART IDENTITY CASCADE;")


@pytest.fixture
def search_engine_prefix():
    engine = SearchEnginePostgreSQL(dsn=TEST_DB_DSN, prefix="my_prefix")
    yield engine
    with engine.conn() as conn:
        conn.execute("TRUNCATE TABLE documents RESTART IDENTITY CASCADE;")


def test_add_document(postgres_service, search_engine):
    content = ["test content"]
    ids = search_engine.add(content)
    assert len(ids) == 1


def test_query_document(postgres_service, search_engine):
    content = ["test query content"]
    search_engine.add(content)
    results = search_engine.query("content")
    assert len(results) == 1


def test_update_document(postgres_service, search_engine):
    content = ["initial content"]
    ids = search_engine.add(content)
    updated_content = ["updated content"]
    search_engine.update(ids, updated_content)
    results = search_engine.query("updated")
    assert len(results) == 1


def test_delete_document(postgres_service, search_engine):
    content = ["content to delete"]
    ids = search_engine.add(content)
    search_engine.delete(ids)
    results = search_engine.query("delete")
    assert len(results) == 0


def test_add(postgres_service, search_engine):
    assert search_engine.query("Lorem") == []
    ids1 = search_engine.add(["Lorem ipsum dolor"])
    ids2 = search_engine.add(["sit amet"])
    assert len(search_engine.query("Lorem")) == 1
    assert search_engine.query("Lorem")[0]["id"] == ids1[0]


def test_query_wildcard(postgres_service, search_engine):
    assert search_engine.query("Lorem") == []
    ids1 = search_engine.add(["Lorem ipsum dolor"])
    ids2 = search_engine.add(["sit amet"])
    assert len(search_engine.query("am*")) == 1
    assert search_engine.query("am*")[0]["id"] == ids2[0]
    assert len(search_engine.query("ame*")) == 1


def test_query_pr(postgres_service, search_engine):
    assert search_engine.query("Lorem") == []
    ids1 = search_engine.add(["Lorem ipsum dolor"])
    ids2 = search_engine.add(["sit amet"])
    assert len(search_engine.query("Lorem or amet")) == 2


def test_add_prefix(postgres_service, search_engine, search_engine_prefix):
    assert search_engine_prefix.query("Lorem") == []
    search_engine_prefix.add(["Lorem ipsum dolor"])
    assert len(search_engine_prefix.query("Lorem")) == 1
    assert len(search_engine.query("Lorem")) == 0


def test_add_id(postgres_service, search_engine):
    ids = search_engine.add(["x"])
    assert len(ids) == 1
    assert len(ids[0]) == 36  # is UUIDv4
    ids = search_engine.add(["y"], ids=["my_id"])
    assert ids == ["my_id"]
    res = search_engine.query("y")
    assert len(res) == 1
    assert res[0]["id"] == "my_id"
    with pytest.raises(psycopg2.errors.UniqueViolation):
        # ID must be unique
        search_engine.add(["z"], ids=["my_id"])


def test_update(postgres_service, search_engine):
    ids = search_engine.add(["Lorem ipsum"])
    res = search_engine.query("Lorem")
    assert len(res) == 1
    assert res[0]["id"] == ids[0]
    search_engine.update(ids=ids, contents=["dolor sit"])
    res = search_engine.query("Lorem")
    assert len(res) == 0
    res = search_engine.query("sit")
    assert len(res) == 1
    assert res[0]["id"] == ids[0]


def test_delete(postgres_service, search_engine):
    ids = search_engine.add(["Lorem ipsum"])
    res = search_engine.query("Lorem")
    assert len(res) == 1
    search_engine.delete(ids)
    res = search_engine.query("Lorem")
    assert len(res) == 0
    search_engine.delete(ids)


def test_query_multiple(postgres_service, search_engine):
    search_engine.add(["Lorem ipsum dolor"])
    search_engine.add(["sit amet"])
    assert len(search_engine.query("Lorem ipsum")) == 1
    assert len(search_engine.query("sit amet")) == 1
    assert len(search_engine.query("Lorem sit")) == 0



from sifts.core import QueryParser


def test_trim_sqlite():
    query = " Lorem\t"
    assert str(QueryParser(query)) == "Lorem"


def test_and_sqlite():
    query = "Lorem and ipsum"
    assert str(QueryParser(query)) == "Lorem AND ipsum"


def test_or_sqlite():
    query = "Lorem or ipsum"
    assert str(QueryParser(query)) == "Lorem OR ipsum"


def test_wildcard_sqlite():
    query = "Lor*"
    assert str(QueryParser(query)) == "Lor*"


def test_wildcard_and_sqlite():
    query = "Lor* and ips*"
    assert str(QueryParser(query)) == "Lor* AND ips*"


def test_trim_postgres():
    query = " Lorem\t"
    assert str(QueryParser(query, backend="postgresql")) == "Lorem"


def test_and_postgres():
    query = "Lorem and ipsum"
    assert str(QueryParser(query, backend="postgresql")) == "Lorem & ipsum"


def test_or_postgres():
    query = "Lorem or ipsum"
    assert str(QueryParser(query, backend="postgresql")) == "Lorem | ipsum"


def test_wildcard_postgres():
    query = "Lor*"
    assert str(QueryParser(query, backend="postgresql")) == "Lor:*"


def test_wildcard_and_postgres():
    query = "Lor* and ips*"
    assert str(QueryParser(query, backend="postgresql")) == "Lor:* & ips:*"



"""Unit tests."""

import os
import sqlite3

import pytest
from sifts.core import SearchEngineSQLite


def test_init(tmp_path):
    path = tmp_path / "search_engine.db"
    SearchEngineSQLite(path)
    assert os.path.isfile(path)
    conn = sqlite3.connect(path)
    cursor = conn.cursor()
    cursor.execute(
        "SELECT name FROM sqlite_master WHERE type='table' AND name='documents';"
    )
    assert cursor.fetchone() is not None
    cursor.execute(
        "SELECT name FROM sqlite_master WHERE type='table' AND name='documents_fts';"
    )
    assert cursor.fetchone() is not None


def test_add(tmp_path):
    path = tmp_path / "search_engine.db"
    search = SearchEngineSQLite(path)
    assert search.query("Lorem") == []
    ids1 = search.add(["Lorem ipsum dolor"])
    ids2 = search.add(["sit amet"])
    assert len(search.query("Lorem")) == 1
    assert search.query("Lorem")[0]["id"] == ids1[0]
    assert len(search.query("am*")) == 1
    assert search.query("am*")[0]["id"] == ids2[0]
    assert len(search.query("Lorem or amet")) == 2


def test_query_multiple(tmp_path):
    path = tmp_path / "search_engine.db"
    search = SearchEngineSQLite(path)
    search.add(["Lorem ipsum dolor"])
    search.add(["sit amet"])
    assert len(search.query("Lorem ipsum")) == 1
    assert len(search.query("sit amet")) == 1
    assert len(search.query("Lorem sit")) == 0


def test_add_prefix(tmp_path):
    path = tmp_path / "search_engine.db"
    search = SearchEngineSQLite(path, prefix="my_prefix")
    assert search.query("Lorem") == []
    search.add(["Lorem ipsum dolor"])
    assert len(search.query("Lorem")) == 1
    search = SearchEngineSQLite(path)
    assert len(search.query("Lorem")) == 0
    search = SearchEngineSQLite(path, prefix="my_prefix")
    assert len(search.query("Lorem")) == 1


def test_add_id(tmp_path):
    path = tmp_path / "search_engine.db"
    search = SearchEngineSQLite(path)
    ids = search.add(["x"])
    assert len(ids) == 1
    assert len(ids[0]) == 36  # is UUIDv4
    ids = search.add(["y"], ids=["my_id"])
    assert ids == ["my_id"]
    res = search.query("y")
    assert len(res) == 1
    assert res[0]["id"] == "my_id"
    with pytest.raises(sqlite3.IntegrityError):
        # ID must be unique
        search.add(["z"], ids=["my_id"])


def test_update(tmp_path):
    path = tmp_path / "search_engine.db"
    search = SearchEngineSQLite(path)
    ids = search.add(["Lorem ipsum"])
    res = search.query("Lorem")
    assert len(res) == 1
    assert res[0]["id"] == ids[0]
    search.update(ids=ids, contents=["dolor sit"])
    res = search.query("Lorem")
    assert len(res) == 0
    res = search.query("sit")
    assert len(res) == 1
    assert res[0]["id"] == ids[0]


def test_delete(tmp_path):
    path = tmp_path / "search_engine.db"
    search = SearchEngineSQLite(path)
    ids = search.add(["Lorem ipsum"])
    res = search.query("Lorem")
    assert len(res) == 1
    search.delete(ids)
    res = search.query("Lorem")
    assert len(res) == 0
    search.delete(ids)


def test_query_metadata(tmp_path):
    path = tmp_path / "search_engine.db"
    search = SearchEngineSQLite(path)
    search.add(["Lorem ipsum dolor"], metadatas=[{"foo": "bar"}])
    search.add(["sit amet"])
    res = search.query("Lorem")
    assert len(res) == 1
    assert res[0]["metadata"] == {"foo": "bar"}
    res = search.query("sit")
    assert len(res) == 1
    assert res[0]["metadata"] is None



"""Core classes for Sifts."""

from __future__ import annotations
import json
import re
import sqlite3
import uuid

import psycopg2
from urllib.parse import urlparse
from contextlib import contextmanager


def make_id():
    return str(uuid.uuid4())


class QueryParser:
    """Parser for search queries."""

    def __init__(self, query: str, backend: str = "sqlite") -> None:
        """Initialize given a query."""
        self.query = query.strip()
        self.backend = backend

    def _to_sqlite(self) -> str:
        query = self.query
        query = re.sub(r"\band\b", "AND", query, flags=re.IGNORECASE)
        query = re.sub(r"\bor\b", "OR", query, flags=re.IGNORECASE)
        return query

    def _to_pg(self) -> str:
        query = self.query

        operators = {"&", "|", "and", "or"}
        words = query.split()
        query_list = []
        i = 0
        while i < len(words):
            if words[i] in operators:
                query_list.append(words[i])
                i += 1
            else:
                query_list.append(words[i])
                if i + 1 < len(words) and words[i + 1] not in operators:
                    query_list.append("&")
                i += 1
        query = " ".join(query_list)
        query = re.sub(r"\band\b", "&", query, flags=re.IGNORECASE)
        query = re.sub(r"\bor\b", "|", query, flags=re.IGNORECASE)
        query = re.sub(r"\b(\w+)\*(?=\s|$|[^\w])", r"\1:*", query)
        return query

    def __str__(self) -> str:
        """Return the right string representation for the backend."""
        if self.backend == "sqlite":
            return self._to_sqlite()
        return self._to_pg()


class SearchEngineBase:

    IS_POSTGRES = False
    QUERY_CREATE_INDEX = ""
    QUERY_CREATE_DOCUMENT = ""
    QUERY_INSERT_DOC = ""
    QUERY_INSERT_INDEX = ""
    QUERY_UPDATE_DOC = ""
    QUERY_UPDATE_INDEX = ""
    QUERY_DELETE = ""
    QUERY_SEARCH = ""
    QUERY_FILTER_META = ""
    QUERY_ORDER_META = ""

    def __init__(self, prefix: str | None = None) -> None:
        self.prefix = prefix
        self.create_tables()

    @contextmanager
    def conn(self):
        """Provide a transactional scope around a series of operations."""
        raise NotImplementedError

    def create_tables(self) -> None:
        with self.conn() as conn:
            conn.execute(self.QUERY_CREATE_INDEX)
            if self.QUERY_CREATE_DOCUMENT:
                conn.execute(self.QUERY_CREATE_DOCUMENT)
            conn.execute("CREATE INDEX IF NOT EXISTS prefix_idx ON documents (prefix)")

    def add(
        self,
        contents: list[str],
        ids: list[str | None] | None = None,
        metadatas: list[dict[str, str] | None] | None = None,
    ) -> int:
        if ids is None:
            ids = [make_id() for _ in contents]
        else:
            ids = [i or make_id() for i in ids]
        if metadatas is None:
            metadatas = [None for _ in contents]
        else:
            metadatas = [json.dumps(m) if m else None for m in metadatas]
        prefixes = [self.prefix for _ in contents]
        with self.conn() as conn:
            conn.executemany(
                self.QUERY_INSERT_DOC, list(zip(contents, ids, metadatas, prefixes))
            )
            if self.IS_POSTGRES:
                conn.executemany(self.QUERY_INSERT_INDEX, [(did,) for did in ids])
            else:
                conn.executemany(self.QUERY_INSERT_INDEX, list(zip(contents, ids)))
        return ids

    def update(
        self,
        ids: list[str],
        contents: list[str],
        metadatas: list[dict[str, str] | None] | None = None,
    ) -> None:
        if metadatas is None:
            metadatas = [None for _ in contents]
        else:
            metadatas = [json.dumps(m) if m else None for m in metadatas]
        with self.conn() as conn:
            if self.IS_POSTGRES:
                conn.executemany(
                    self.QUERY_UPDATE_INDEX, list(zip(contents, contents, ids))
                )
            else:
                conn.executemany(self.QUERY_UPDATE_INDEX, list(zip(contents, ids)))
            if metadatas and any(metadatas):
                metadatas = [json.dumps(m) if m else None for m in metadatas]
                conn.executemany(
                    self.QUERY_UPDATE_DOC,
                    list(zip(metadatas, ids)),
                )

    def delete(self, ids: list[str]) -> None:
        with self.conn() as conn:
            conn.executemany(self.QUERY_DELETE_INDEX, (ids,))
            conn.executemany(self.QUERY_DELETE_DOC, (ids,))

    def query(
        self,
        query_string: str,
        limit: int = 0,
        where: dict | None = None,
        order_by: str | None = None,
        descending: bool = False,
    ) -> list:
        with self.conn() as conn:
            try:
                fts_query = self.QUERY_SEARCH
                if self.prefix is None:
                    fts_query += " AND prefix IS NULL"
                else:
                    fts_query += f" AND prefix = '{self.prefix}'"

                backend = "postgresql" if self.IS_POSTGRES else "sqlite"
                query_string = str(QueryParser(query_string, backend=backend))
                params = [query_string]

                if where:
                    for key, value in where.items():
                        fts_query += " AND " + self.QUERY_FILTER_META.format(key)
                        params.append(value)

                if limit:
                    fts_query += " LIMIT (?)"

                if order_by:
                    fts_query += " ORDER BY " + self.QUERY_ORDER_META.format(order_by)
                    if descending:
                        fts_query += " DESC"
                result = conn.execute(fts_query, params) or []
                if self.IS_POSTGRES:
                    result = conn.fetchall()
                result = [
                    {
                        "id": match[0],
                        "rank": match[1],
                        "content": match[2],
                        "metadata": json.loads(match[3] or "null"),
                    }
                    for match in result
                ]
            except (sqlite3.OperationalError, psycopg2.OperationalError):
                return []
        return result


class SearchEngineSQLite(SearchEngineBase):

    QUERY_CREATE_INDEX = (
        "CREATE VIRTUAL TABLE IF NOT EXISTS documents_fts USING fts5(id, content)"
    )
    QUERY_CREATE_DOCUMENT = """CREATE TABLE IF NOT EXISTS documents (
            id TEXT PRIMARY KEY,
            content TEXT,
            prefix TEXT,
            metadata JSON
        )"""
    QUERY_INSERT_INDEX = "INSERT INTO documents_fts (content, id) VALUES (?, ?)"
    QUERY_INSERT_DOC = (
        "INSERT INTO documents (content, id, metadata, prefix) VALUES (?, ?, ?, ?)"
    )
    QUERY_UPDATE_INDEX = "UPDATE documents_fts SET content = (?) WHERE id = (?)"
    QUERY_UPDATE_DOC = "UPDATE documents SET metadata = (?) WHERE id = (?)"
    QUERY_DELETE_INDEX = "DELETE FROM documents_fts WHERE id = (?)"
    QUERY_DELETE_DOC = "DELETE FROM documents WHERE id = (?)"
    QUERY_SEARCH = """SELECT doc.id, fts.rank, doc.content, doc.metadata
                FROM documents_fts fts
                JOIN documents doc ON doc.id = fts.id
                WHERE fts.content MATCH (?)
                """
    QUERY_FILTER_META = 'json_extract(doc.metadata, "$.{}") = (?)'
    QUERY_ORDER_META = 'json_extract(doc.metadata, "$.{}")'

    def __init__(self, db_path="search_engine.db", prefix: str | None = None) -> None:
        self.db_path = db_path
        super().__init__(prefix=prefix)

    @contextmanager
    def conn(self):
        """Provide a transactional scope around a series of operations."""
        conn = sqlite3.connect(self.db_path)
        try:
            yield conn
            conn.commit()
        finally:
            conn.close()


class SearchEnginePostgreSQL(SearchEngineBase):

    IS_POSTGRES = True
    QUERY_CREATE_INDEX = """
        CREATE TABLE IF NOT EXISTS documents (
            id TEXT PRIMARY KEY,
            content TEXT,
            prefix TEXT,
            metadata JSONB,
            tsvector TSVECTOR
        );
        CREATE INDEX IF NOT EXISTS documents_tsvector_idx ON documents USING GIN (tsvector);
        CREATE INDEX IF NOT EXISTS prefix_idx ON documents (prefix);
        """

    QUERY_CREATE_DOCUMENT = ""

    QUERY_INSERT_INDEX = (
        "UPDATE documents SET tsvector = to_tsvector(content) WHERE id = %s"
    )
    QUERY_INSERT_DOC = (
        "INSERT INTO documents (content, id, metadata, prefix) VALUES (%s, %s, %s, %s)"
    )

    QUERY_UPDATE_INDEX = "UPDATE documents SET content = %s, tsvector = to_tsvector('simple', %s) WHERE id = %s"
    QUERY_UPDATE_DOC = "UPDATE documents SET metadata = %s WHERE id = %s"

    QUERY_DELETE_INDEX = "UPDATE documents SET tsvector = NULL WHERE id = %s"
    QUERY_DELETE_DOC = "DELETE FROM documents WHERE id = %s"

    QUERY_SEARCH = """
    SELECT id, ts_rank(tsvector, query) AS rank, content, metadata
    FROM documents, to_tsquery('simple', %s) query
    WHERE tsvector @@ query
    """

    QUERY_FILTER_META = "metadata->>%s = %s"
    QUERY_ORDER_META = "metadata->>%s"

    def __init__(
        self,
        dsn,
        prefix: str | None = None,
    ) -> None:
        self.dsn = dsn
        super().__init__(prefix=prefix)

    @contextmanager
    def conn(self):
        """Provide a transactional scope around a series of operations."""
        conn = psycopg2.connect(dsn=self.dsn)
        try:
            yield conn.cursor()
            conn.commit()
        finally:
            conn.close()


def db_url_to_dsn(db_url: str) -> str:
    """Convert a database URL to a DSN"""
    url = urlparse(db_url)
    dbname = url.path[1:]
    user = url.username
    password = url.password
    host = url.hostname
    port = url.port
    dsn = f"dbname={dbname} user={user} password={password} host={host} port={port}"
    return dsn


def SearchEngine(db_url: str, prefix: str | None = None) -> SearchEngineBase:
    """Constructor for search engine instance."""
    if not db_url:
        return SearchEngineSQLite(prefix=prefix)
    if db_url.startswith("sqlite:///"):
        return SearchEngineSQLite(db_path=db_url[10:], prefix=prefix)
    return SearchEnginePostgreSQL(dsn=db_url_to_dsn(db_url))
